"""SSH service module."""

from __future__ import annotations

from ubo_gui.menu.types import (
    ActionItem,
    ApplicationItem,
    HeadedMenu,
    HeadlessMenu,
    SubMenuItem,
)
from ubo_gui.prompt import PromptWidget

from ubo_app.store import dispatch
from ubo_app.store.main import RegisterSettingAppAction
from ubo_app.store.services.notifications import (
    Importance,
    Notification,
    NotificationDisplayType,
    NotificationsAddAction,
)
from ubo_app.utils.server import send_command


class ClearTemporaryUsersPrompt(PromptWidget):
    """Prompt to clear all temporary users."""

    def first_option_callback(self: ClearTemporaryUsersPrompt) -> None:
        """Clear all temporary users."""
        self.dispatch('on_close')

    def second_option_callback(self: ClearTemporaryUsersPrompt) -> None:
        """Close the prompt."""
        send_command('ssh clear_all_temporary_accounts')
        self.dispatch('on_close')
        dispatch(
            NotificationsAddAction(
                notification=Notification(
                    title='All SSH Accounts Removed',
                    content='All SSH accounts have been removed.',
                    importance=Importance.MEDIUM,
                    icon='󰣀',
                    display_type=NotificationDisplayType.FLASH,
                ),
            ),
        )

    def __init__(self: ClearTemporaryUsersPrompt, **kwargs: object) -> None:
        """Initialize the prompt."""
        super().__init__(**kwargs, items=None)
        self.prompt = 'Clear temporary ssh users?'
        self.icon = '󱈋'
        self.first_option_label = ''
        self.first_option_background_color = 'black'
        self.first_option_is_short = False
        self.second_option_label = 'Delete'
        self.second_option_icon = '󱈊'
        self.second_option_is_short = False


def init_service() -> None:
    """Initialize the SSH service."""

    def create_ssh_account() -> None:
        result = send_command('ssh create_temporary_ssh_account', has_output=True)
        username, password = result.split(':')
        dispatch(
            NotificationsAddAction(
                notification=Notification(
                    title='Temporary SSH Account Created',
                    content=f"""Username: {username}
Password: {password}
Make sure to delete it after use. Note that in order to make things work for you, we \
had to make sure password authentication for ssh server is enabled, you may want to \
disable it later.""",
                    importance=Importance.MEDIUM,
                    icon='󰣀',
                    display_type=NotificationDisplayType.STICKY,
                ),
            ),
        )

    dispatch(
        RegisterSettingAppAction(
            menu_item=SubMenuItem(
                label='SSH',
                icon='󰣀',
                sub_menu=HeadlessMenu(
                    title='SSH',
                    items=[
                        SubMenuItem(
                            label='Setup',
                            icon='',
                            sub_menu=HeadedMenu(
                                title='SSH Setup',
                                heading='Create an SSH account',
                                sub_heading='This will create a temporary SSH account,'
                                ' you should delete it after use.',
                                items=[
                                    ActionItem(
                                        label='Create',
                                        icon='󰗹',
                                        action=create_ssh_account,
                                    ),
                                ],
                            ),
                        ),
                        ApplicationItem(
                            label='Remove all users',
                            icon='󱈊',
                            application=ClearTemporaryUsersPrompt,
                        ),
                    ],
                ),
            ),
        ),
    )



# ruff: noqa: D100, D101, D102, D103, D104, D107
from __future__ import annotations

import os
import sys
import threading
import traceback
from pathlib import Path
from typing import TYPE_CHECKING

import dotenv
import sentry_sdk
from redux import FinishAction

if TYPE_CHECKING:
    from types import TracebackType

dotenv.load_dotenv(Path(__file__).parent / '.dev.env')


def setup_logging() -> None:
    from ubo_app.constants import GUI_LOG_LEVEL, LOG_LEVEL

    if LOG_LEVEL:
        import logging

        import ubo_app.logging

        level = getattr(
            ubo_app.logging,
            LOG_LEVEL,
            getattr(logging, LOG_LEVEL, logging.INFO),
        )

        ubo_app.logging.logger.setLevel(level)
        ubo_app.logging.add_file_handler(ubo_app.logging.logger, level)
        ubo_app.logging.add_stdout_handler(ubo_app.logging.logger, level)
    if GUI_LOG_LEVEL:
        import logging

        import ubo_gui.logger

        level = getattr(
            ubo_gui.logger,
            GUI_LOG_LEVEL,
            getattr(logging, GUI_LOG_LEVEL, logging.INFO),
        )

        ubo_gui.logger.logger.setLevel(level)
        ubo_gui.logger.add_file_handler(level)
        ubo_gui.logger.add_stdout_handler(level)


def setup_sentry() -> None:  # pragma: no cover
    from ubo_app.constants import SENTRY_DSN

    if SENTRY_DSN:
        sentry_sdk.init(
            dsn=SENTRY_DSN,
            traces_sample_rate=1.0,
            profiles_sample_rate=1.0,
        )


def get_all_thread_stacks() -> dict[str, list[str]]:
    id_to_name = {th.ident: th.name for th in threading.enumerate()}
    thread_stacks = {}
    for thread_id, frame in sys._current_frames().items():  # noqa: SLF001
        thread_stacks[id_to_name.get(thread_id, f'unknown-{thread_id}')] = (
            traceback.format_stack(frame)
        )
    return thread_stacks


def global_exception_handler(
    exception_type: type[BaseException],
    exception_value: BaseException,
    exception_traceback: TracebackType,
) -> None:
    from ubo_app.logging import logger

    error_message = ''.join(
        traceback.format_exception(
            exception_type,
            exception_value,
            exception_traceback,
        ),
    )
    threads_info = get_all_thread_stacks()

    logger.debug(
        f'Uncaught exception: {exception_type}: {exception_value}\n{error_message}',
        extra={'threads': threads_info},
    )


def main() -> None:
    """Instantiate the `MenuApp` and run it."""
    os.environ['KIVY_NO_CONFIG'] = '1'
    os.environ['KIVY_NO_FILELOG'] = '1'
    os.environ['KIVY_NO_CONSOLELOG'] = '1'
    os.environ['KCFG_KIVY_EXIT_ON_ESCAPE'] = '0'

    setup_sentry()
    setup_logging()

    # Set the global exception handler
    sys.excepthook = global_exception_handler
    threading.excepthook = global_exception_handler

    if len(sys.argv) > 1 and sys.argv[1] == 'bootstrap':
        from ubo_app.system.bootstrap import bootstrap

        bootstrap(
            with_docker='--with-docker' in sys.argv,
            for_packer='--for-packer' in sys.argv,
        )
        sys.exit(0)

    import headless_kivy_pi.config

    headless_kivy_pi.config.setup_headless_kivy({'automatic_fps': True})

    from kivy.clock import Clock

    from ubo_app.load_services import load_services
    from ubo_app.menu import MenuApp

    load_services()
    app = MenuApp()

    try:
        app.run()
    finally:
        from ubo_app.store import dispatch

        dispatch(FinishAction())

        # Needed since redux is scheduled using Clock scheduler and Clock doesn't run
        # after app is stopped.
        Clock.tick()


if __name__ == '__main__':
    main()



"""Module for sending commands to the system manager socket."""

from __future__ import annotations

import socket
import threading
from typing import Literal, overload

from ubo_app.constants import SOCKET_PATH
from ubo_app.logging import logger

thread_lock = threading.Lock()


@overload
def send_command(command: str) -> None: ...


@overload
def send_command(command: str, *, has_output: Literal[True]) -> str: ...


def send_command(command: str, *, has_output: bool = False) -> str | None:
    """Send a command to the system manager socket."""
    client = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    try:
        client.connect(SOCKET_PATH)
    except Exception as exception:  # noqa: BLE001
        logger.error('Unable to connect to the socket', exc_info=exception)
        if has_output:
            return ''
        return None

    output = None
    with thread_lock:
        client.sendall(f'{command}'.encode() + b'\0')
        remaining = b''
        while has_output:
            datagram = remaining + client.recv(1024)
            if not datagram:
                break
            if b'\0' not in datagram:
                remaining = datagram
                continue
            response, remaining = datagram.split(b'\0', 1)
            output = response.decode('utf-8')
            logger.debug('Server response:', extra={'response': response})
        client.close()

    return output



"""provides a function to install and start Docker on the host machine."""

from __future__ import annotations

import subprocess
from pathlib import Path


def ssh_handler(command: str) -> str | None:
    """Install and start Docker on the host machine."""
    if command == 'create_temporary_ssh_account':
        result = subprocess.run(
            Path(__file__).parent.joinpath('create_temporary_ssh_account.sh'),  # noqa: S603
            check=True,
            text=True,
            stdout=subprocess.PIPE,
        )
        result.check_returncode()
        return result.stdout
    if command == 'clear_all_temporary_accounts':
        subprocess.run(
            Path(__file__).parent.joinpath('clear_all_temporary_accounts.sh'),  # noqa: S603
            check=False,
        )
    return None
